/*
 * Simple program to grab images from VDR Recording
 *
 * Copyright (c) 2005 Andreas Brachold
 * 
 * based on FFmpeg main Copyright (c) 2000-2003 Fabrice Bellard
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include <stdio.h>
#include <unistd.h>
#include <errno.h>

#include <iostream>
#include <string>
#include <vector>
#include "ffm.h"

# include <avcodec.h>
# include <avformat.h>

static int frame_width = 160;
static int frame_height = 120;
static float frame_aspect_ratio = 0;
static enum PixelFormat frame_pix_fmt = PIX_FMT_YUV420P;
static int frame_rate = 25;
static int frame_rate_base = 1;
static int keep_aspect_ratio = 1;



typedef struct AVOutputStream
{
	int index;		/* stream index in the output file */
	int source_index;	/* AVInputStream index */
	AVStream *st;		/* stream in the output file */
	int frame_number;
	/* video only */
	int video_resample;	/* video_resample */
	AVPicture pict_tmp;	/* temporary image for resampling */
	ImgReSampleContext *img_resample_ctx;	/* for image resampling */
} AVOutputStream;

typedef struct AVInputStream
{
	int index;
	AVStream *st;
	int discard;		/* true if stream data should be discarded */
	unsigned long frame;	/* current frame */
} AVInputStream;

static inline int 
do_write_frame(AVFormatContext *s, 
               int stream_index, 
               const uint8_t *buf, 
               int size)
{
#if FFMPEG_VERSION_INT == 0x000408
    return av_write_frame (s, stream_index,buf,size);
/*#elif FFMPEG_VERSION_INT == 0x000409
    AVPacket pkt;
    av_init_packet(&pkt);
    pkt.stream_index= stream_index;
    pkt.data= (uint8_t *)buf;
    pkt.size= size;
    return av_interleaved_write_frame(s, &pkt);
*/
#else
    #error "there is a current unsupported ffmpeg version or FFMPEG_VERSION_INT isnt defined"
#endif
}

static void
do_video_out (AVFormatContext * s,
	      AVOutputStream * ost,
	      AVInputStream * ist, AVPicture * in_picture, int *frame_size)
{
	int ret;
	AVPicture *final_picture, *formatted_picture;
	AVPicture picture_format_temp;
	static uint8_t *video_buffer;
	uint8_t *buf = NULL;
	AVCodecContext *enc, *dec;

#define VIDEO_BUFFER_SIZE (1024*1024)

	enc = &ost->st->codec;
	dec = &ist->st->codec;

	*frame_size = 0;

	if (!video_buffer)
		video_buffer = (uint8_t *) av_malloc (VIDEO_BUFFER_SIZE);
	if (!video_buffer)
		return;

	/* convert pixel format if needed */
	if (enc->pix_fmt != dec->pix_fmt)
	{
		int size;

		/* create temporary picture */
		size = avpicture_get_size (enc->pix_fmt, dec->width,
					   dec->height);
		buf = (uint8_t *) av_malloc (size);
		if (!buf)
			return;
		formatted_picture = &picture_format_temp;
		avpicture_fill (formatted_picture, buf, enc->pix_fmt,
				dec->width, dec->height);

		if (img_convert (formatted_picture, enc->pix_fmt,
				 in_picture, dec->pix_fmt,
				 dec->width, dec->height) < 0)
		{
			fprintf (stderr,
				 "pixel format conversion not handled\n");
			goto the_end;
		}
	}
	else
	{
		formatted_picture = in_picture;
	}

	/* XXX: resampling could be done before raw format convertion in
	 * some cases to go faster */
	/* XXX: only works for YUV420P */
	if (ost->video_resample)
	{
		final_picture = &ost->pict_tmp;
		img_resample (ost->img_resample_ctx, final_picture,
			      formatted_picture);
	}
	else
	{
		final_picture = formatted_picture;
	}

	if (s->oformat->flags & AVFMT_RAWPICTURE)
	{
		/* raw pictures are written as AVPicture structure to
		 * avoid any copies. We support temorarily the older
		 * method. */
		do_write_frame (s, ost->index,
				(uint8_t *) final_picture,
				sizeof (AVPicture));
	}
	else
	{
		AVFrame big_picture;

		memset (&big_picture, 0, sizeof (AVFrame));
		*(AVPicture *) & big_picture = *final_picture;
		big_picture.quality = ost->st->quality;

		ret = avcodec_encode_video (enc,
					    video_buffer,
					    VIDEO_BUFFER_SIZE, &big_picture);
		//enc->frame_number = enc->real_pict_num;
		do_write_frame (s, ost->index, video_buffer, ret);
		*frame_size = ret;
		//fprintf(stderr,"\nFrame: %3d %3d size: %5d type: %d",
		//        enc->frame_number-1, enc->real_pict_num, ret,
		//        enc->pict_type);
	}
	ost->frame_number++;

      the_end:
	av_free (buf);
}

/*
 * The following code is the main loop of the file converter
 */
static bool
av_encode (AVFormatContext * ofc,
	   AVFormatContext * ifc, const tPackedList & packed)
{
	int ret, i, j, k, n;
	AVCodecContext *codec, *icodec;
	AVOutputStream *ost, **ost_table = NULL;
	AVInputStream *ist, **ist_table = NULL;
	tPackedList::const_iterator ti = packed.begin ();
	tPackedList::const_iterator te = packed.end ();
	int frame;


	/* input stream init */
	ist_table =
		(AVInputStream **) av_mallocz (ifc->nb_streams *
					       sizeof (AVInputStream *));
	if (!ist_table)
		goto fail;

	for (i = 0; i < ifc->nb_streams; ++i)
	{
		ist = (AVInputStream *) av_mallocz (sizeof (AVInputStream));
		if (!ist)
			goto fail;
		ist_table[i] = ist;
	}
	j = 0;
	for (k = 0; k < ifc->nb_streams; ++k)
	{
		ist = ist_table[j++];
		ist->st = ifc->streams[k];
		ist->index = k;
		ist->discard = 1;	/* the stream is discarded by default
					 * (changed later) */
	}

	/* output stream init */
	ost_table =
		(AVOutputStream **) av_mallocz (sizeof (AVOutputStream *) *
						ofc->nb_streams);
	if (!ost_table)
		goto fail;
	for (i = 0; i < ofc->nb_streams; ++i)
	{
		ost = (AVOutputStream *) av_mallocz (sizeof (AVOutputStream));
		if (!ost)
			goto fail;
		ost_table[i] = ost;
	}

	n = 0;
	for (i = 0; i < ofc->nb_streams; ++i)
	{
		int found;
		ost = ost_table[n++];
		ost->index = i;
		ost->st = ofc->streams[i];
		/* get corresponding input stream index : we select the first one with the right type */
		found = 0;
		for (j = 0; j < ifc->nb_streams; ++j)
		{
			ist = ist_table[j];
			if (ist->discard &&
			    ist->st->codec.codec_type ==
			    ost->st->codec.codec_type)
			{
				ost->source_index = j;
				found = 1;
			}
		}

		if (!found)
		{
			/* try again and reuse existing stream */
			for (j = 0; j < ifc->nb_streams; ++j)
			{
				ist = ist_table[j];
				if (ist->st->codec.
				    codec_type == ost->st->codec.codec_type)
				{
					ost->source_index = j;
					found = 1;
				}
			}
			if (!found)
			{
				fprintf (stderr,
					 "Could not find input stream matching output stream %d\n",
					 ost->index);
				return false;
			}
		}
		ist = ist_table[ost->source_index];
		ist->discard = 0;
	}


	/* for each output stream, we compute the right encoding parameters */
	for (i = 0; i < ofc->nb_streams; ++i)
	{
		ost = ost_table[i];
		ist = ist_table[ost->source_index];

		codec = &ost->st->codec;
		icodec = &ist->st->codec;


		if (codec->codec_type == CODEC_TYPE_VIDEO)
		{
			if (codec->width == icodec->width &&
			    codec->height == icodec->height)
			{
				ost->video_resample = 0;
			}
			else
			{
				uint8_t *buf;
				ost->video_resample = 1;
				buf = (uint8_t *)
					av_malloc ((codec->width *
						    codec->height * 3) / 2);
				if (!buf)
					goto fail;
				ost->pict_tmp.data[0] = buf;
				ost->pict_tmp.data[1] =
					ost->pict_tmp.data[0] +
					(codec->width * codec->height);
				ost->pict_tmp.data[2] =
					ost->pict_tmp.data[1] +
					(codec->width * codec->height) / 4;
				ost->pict_tmp.linesize[0] = codec->width;
				ost->pict_tmp.linesize[1] = codec->width / 2;
				ost->pict_tmp.linesize[2] = codec->width / 2;

				ost->img_resample_ctx =
#if FFMPEG_VERSION_INT == 0x000408
					img_resample_full_init (ost->st->codec.width,
								ost->st->codec.height,
                                ist->st->codec.width,
								ist->st->codec.height,
								0, 0, 0, 0
                                );
/*#elif FFMPEG_VERSION_INT == 0x000409
					img_resample_full_init (ost->st->codec.width,
								ost->st->codec.height,
                                ist->st->codec.width,
								ist->st->codec.height,
								0, 0, 0, 0,
								0, 0, 0, 0
                                );
*/
#else
    #error "there is a current unsupported ffmpeg version or FFMPEG_VERSION_INT isnt defined"
#endif
			}
		}
		else
		{
			fprintf (stderr,
				 "Unsupported codec_type for output stream %d\n",
				 ost->index);
			return false;
		}

	}

#ifdef DEBUG
	/* dump the file output parameters - cannot be done before in case
	 * of stream copy */
	dump_format (ofc, 0, ofc->filename, 1);
#endif

	/* open each encoder */
	for (i = 0; i < ofc->nb_streams; ++i)
	{
		ost = ost_table[i];
		AVCodec *codec;
		codec = avcodec_find_encoder (ost->st->codec.codec_id);
		if (!codec)
		{
			fprintf (stderr,
				 "Unsupported codec for output stream %d\n",
				 ost->index);
			return false;
		}
		if (avcodec_open (&ost->st->codec, codec) < 0)
		{
			fprintf (stderr,
				 "Error while opening codec for stream %d - maybe incorrect parameters such as bit_rate, rate, width or height\n",
				 ost->index);
			return false;
		}
	}

	/* open each decoder */
	for (i = 0; i < ifc->nb_streams; ++i)
	{
		ist = ist_table[i];
		AVCodec *codec;
		codec = avcodec_find_decoder (ist->st->codec.codec_id);
		if (!codec)
		{
			fprintf (stderr,
				 "Unsupported codec (id=%d) for input stream %d\n",
				 ist->st->codec.codec_id, ist->index);
			return false;
		}
		if (avcodec_open (&ist->st->codec, codec) < 0)
		{
			fprintf (stderr,
				 "Error while opening codec for input stream %d\n",
				 ist->index);
			return false;
		}
	}


	/* open files and write file headers */
	if (av_write_header (ofc) < 0)
	{
		fprintf (stderr,
			 "Could not write header for output file #%d (incorrect codec paramters ?)\n",
			 i);
		ret = -EINVAL;
		goto fail;
	}

	for (frame = 0; ti != te && frame == 0; ++ti)
	{
		uint8_t *ptr;
		int len;
		int /*data_size, */ got_picture;
		AVPicture picture;

		if (false)
		{
		      redo:
			continue;
		}

		/* read a packet from it and output it in the fifo */
		ist = ist_table[0];
		if (ist->discard)
			goto discard_packet;

		len = (*ti).nPackedLength;
		ptr = (*ti).pPacked;
		while (len > 0)
		{
			/* decode the packet if needed */
			/*data_size = 0; */
			switch (ist->st->codec.codec_type)
			{
			case CODEC_TYPE_VIDEO:
			{
				AVFrame big_picture;

				/*data_size =
				 * (ist->st->codec.width *
				 * ist->st->codec.height * 3) / 2; */
				ret = avcodec_decode_video (&ist->st->
							    codec,
							    &big_picture,
							    &got_picture,
							    ptr, len);
				picture = *(AVPicture *) & big_picture;
				ist->st->quality = big_picture.quality;
				if (ret < 0)
				{
				      fail_decode:
					fprintf (stderr,
						 "Error while decoding stream %d\n",
						 ist->index);
					goto redo;
				}
				if (!got_picture)
				{
					/* no picture yet */
					ptr += ret;
					len -= ret;
					continue;
				}
			}
				break;
			default:
				goto fail_decode;
			}
			ptr += ret;
			len -= ret;

			/* transcode raw format, encode packets and output them */

			for (i = 0; i < ofc->nb_streams; ++i)
			{
				int frame_size;

				ost = ost_table[i];
				if (CODEC_TYPE_VIDEO ==
				    ost->st->codec.codec_type)
					/* find an audio stream for synchro */
				{
					do_video_out (ofc, ost,
						      ist,
						      &picture, &frame_size);
					++frame;
				}
				else
				{
					fprintf (stderr,
						 "Unsupported codec_type for output stream %d\n",
						 ost->index);
					return false;
				}
			}
		}
	      discard_packet:
		;
	}

	/* close each encoder */
	for (i = 0; i < ofc->nb_streams; ++i)
	{
		ost = ost_table[i];
		av_freep (&ost->st->codec.stats_in);
		avcodec_close (&ost->st->codec);
	}

	/* close each decoder */
	for (i = 0; i < ifc->nb_streams; ++i)
	{
		ist = ist_table[i];
		avcodec_close (&ist->st->codec);
	}


	/* write the trailer if needed and close file */
	av_write_trailer (ofc);
	/* finished ! */

	ret = 0;
      fail1:
	if (ist_table)
	{
		for (i = 0; i < ifc->nb_streams; ++i)
		{
			ist = ist_table[i];
			av_free (ist);
		}
		av_free (ist_table);
	}
	if (ost_table)
	{
		for (i = 0; i < ofc->nb_streams; ++i)
		{
			ost = ost_table[i];
			if (ost)
			{
				av_free (ost->pict_tmp.data[0]);
				if (ost->video_resample)
					img_resample_close (ost->
							    img_resample_ctx);
				av_free (ost);
			}
		}
		av_free (ost_table);
	}
	return ret;
      fail:
	ret = -ENOMEM;
	goto fail1;
}


void
print_error (const char *filename, int err)
{
	switch (err)
	{
	case AVERROR_NUMEXPECTED:
		fprintf (stderr, "%s: Incorrect image filename syntax.\n"
			 "Use '%%d' to specify the image number:\n"
			 "  for img1.jpg, img2.jpg, ..., use 'img%%d.jpg';\n"
			 "  for img001.jpg, img002.jpg, ..., use 'img%%03d.jpg'.\n",
			 filename);
		break;
	case AVERROR_INVALIDDATA:
		fprintf (stderr, "%s: Error while parsing header\n",
			 filename);
		break;
	case AVERROR_NOFMT:
		fprintf (stderr, "%s: Unknown format\n", filename);
		break;
	default:
		fprintf (stderr, "%s: Error while opening file\n", filename);
		break;
	}
}

/* state for codec information */
#define CSTATE_NOTFOUND    0
#define CSTATE_DECODING    1
#define CSTATE_FOUND       2

static int
has_codec_parameters (AVCodecContext * enc)
{
	int val;
	switch (enc->codec_type)
	{
	case CODEC_TYPE_AUDIO:
		val = enc->sample_rate;
		break;
	case CODEC_TYPE_VIDEO:
		val = enc->width;
		break;
	default:
		val = 1;
		break;
	}
	return (val != 0);
}

/**
 * Read the beginning of a media file to get stream information. This
 * is useful for file formats with no headers such as MPEG. This
 * function also compute the real frame rate in case of mpeg2 repeat
 * frame mode.
 *
 * @param ic media file handle
 * @return >=0 if OK. AVERROR_xxx if error.  
 */
int
my_find_stream_info (AVFormatContext * ic, const tPackedList & packed)
{
	int i, ret, count, got_picture, size, read_size;
	AVCodec *codec;
	AVStream *st;
	AVFrame picture;
	uint8_t *ptr;
	int min_read_size, max_read_size;

	tPackedList::const_iterator ti = packed.begin ();
	tPackedList::const_iterator te = packed.end ();
	ret = -1;
	min_read_size = 250000;
	/* max read size is 2 seconds of video max */
	max_read_size = min_read_size * 10;

	int codec_info_state = CSTATE_NOTFOUND;
	int codec_info_nb_repeat_frames = 0;
	int codec_info_nb_real_frames = 0;

	/* set initial codec state */
	for (i = 0; i < ic->nb_streams; ++i)
	{
		st = ic->streams[i];
		if (has_codec_parameters (&st->codec))
			codec_info_state = CSTATE_FOUND;
	}

	count = 0;
	read_size = 0;

	for (; ti != te; ++ti)
	{
		/* check if one codec still needs to be handled */
		for (i = 0; i < ic->nb_streams; ++i)
		{
			st = ic->streams[i];
			if (codec_info_state != CSTATE_FOUND)
				break;
		}
		if (i == ic->nb_streams)
		{
			/* NOTE: if the format has no header, then we need to read
			 * some packets to get most of the streams, so we cannot
			 * stop here */
			/*if (!(ic->iformat->flags & AVFMT_NOHEADER) ||
			    read_size >= min_read_size)*/
			{
				/* if we found the info for all the codecs, we can stop */
				ret = count;
				break;
			}
		}
		else
		{
			/* we did not get all the codec info, but we read too much data */
			if (read_size >= max_read_size)
			{
				ret = count;
				break;
			}
		}

		read_size += (*ti).nPackedLength;

		/* open new codecs */
		for (i = 0; i < ic->nb_streams; ++i)
		{
			st = ic->streams[i];
			if (codec_info_state == CSTATE_NOTFOUND)
			{
				/* set to found in case of error */
				codec_info_state = CSTATE_FOUND;
				codec = avcodec_find_decoder (st->codec.
							      codec_id);
				if (codec)
				{
					if (codec->
					    capabilities &
					    CODEC_CAP_TRUNCATED)
						st->codec.flags |=
							CODEC_FLAG_TRUNCATED;

					ret = avcodec_open (&st->codec,
							    codec);
					if (ret >= 0)
						codec_info_state =
							CSTATE_DECODING;
				}
			}
		}

		st = ic->streams[0];
		if (codec_info_state == CSTATE_DECODING)
		{
			/* decode the data and update codec parameters */
			size = (*ti).nPackedLength;
			ptr = (*ti).pPacked;
			while (size > 0)
			{
				if (CODEC_TYPE_VIDEO == st->codec.codec_type)
				{
					ret = avcodec_decode_video (&st->
								    codec,
								    &picture,
								    &got_picture,
								    ptr,
								    size);
				}
				else
				{
					ret = -1;
				}
				if (ret < 0)
				{
					/* if error, simply ignore because another packet
					 * may be OK */
					break;
				}
				if (got_picture)
				{
					/* we got the parameters - now we can stop
					 * examining this stream */
					/* XXX: add a codec info so that we can decide if
					 * the codec can repeat frames */
					if (st->codec.codec_id ==
					    CODEC_ID_MPEG1VIDEO
					    && ic->iformat != &mpegts_demux
					    && st->codec.sub_id == 2)
					{
						/* for mpeg2 video, we want to know the real
						 * frame rate, so we decode 40 frames. In mpeg
						 * TS case we do not do it because it would be
						 * too long */
						codec_info_nb_real_frames++;
						codec_info_nb_repeat_frames += st->codec.coded_frame->repeat_pict;

						/* stop after 40 frames */
						if (codec_info_nb_real_frames >= 40)
						{
							av_reduce (&st->r_frame_rate,&st->r_frame_rate_base,
								   (int64_t)st->codec.frame_rate * codec_info_nb_real_frames,
								   (codec_info_nb_real_frames + (codec_info_nb_repeat_frames >> 1)) *
								   st->codec.frame_rate_base,1 << 30);
							goto close_codec;
						}
					}
					else
					{
					      close_codec:
						codec_info_state =
							CSTATE_FOUND;
						avcodec_close (&st->codec);
						break;
					}
				}
				ptr += ret;
				size -= ret;
			}
		}
		count++;
	}

	/* close each codec if there are opened */
	for (i = 0; i < ic->nb_streams; ++i)
	{
		st = ic->streams[i];
		if (codec_info_state == CSTATE_DECODING)
			avcodec_close (&st->codec);
	}

	/* set real frame rate info */
	for (i = 0; i < ic->nb_streams; ++i)
	{
		st = ic->streams[i];
		if (CODEC_TYPE_VIDEO == st->codec.codec_type)
		{
			if (!st->r_frame_rate)
			{
				st->r_frame_rate = st->codec.frame_rate;
				st->r_frame_rate_base =
					st->codec.frame_rate_base;
			}
		}
	}


    return ret;
}

static AVFormatContext *
opt_input_file (const tPackedList & packed)
{
	AVInputFormat *file_iformat;
	AVFormatContext *ic;
	AVFormatParameters params, *ap = &params;
	int err, i, ret, rfps, rfps_base;

	// get default parameters from command line 
	memset (ap, 0, sizeof (*ap));
	ap->frame_rate = frame_rate;
	ap->frame_rate_base = frame_rate_base;
	ap->width = frame_width;
	ap->height = frame_height;
	ap->image_format = NULL;
	ap->pix_fmt = frame_pix_fmt;


	AVProbeData probe_data, *pd = &probe_data;
	pd->filename = "";
	pd->buf = (*packed.begin ()).pPacked;
	pd->buf_size = (*packed.begin ()).nPackedLength;

	file_iformat = av_probe_input_format (pd, 1);

	// open the input file with generic libav function 
	err = av_open_input_file (&ic, "pipe:", file_iformat, 0, ap);
	if (err < 0)
	{
		print_error ("", err);
		return false;
	}

	// If not enough info to get the stream parameters, we decode the
	//   first frames to get it. (used in mpeg case for example) 
	ret = my_find_stream_info (ic, packed);
	if (ret < 0)
	{
		fprintf (stderr,
			 "could not find video codec parameters\n");
		return false;
	}

	// update the current parameters so that they match the one of the input stream 
	for (i = 0; i < ic->nb_streams; ++i)
	{
		AVCodecContext *enc = &ic->streams[i]->codec;
		if (CODEC_TYPE_VIDEO == enc->codec_type)
		{
			frame_height = enc->height;
			frame_width = enc->width;
#if FFMPEG_VERSION_INT == 0x000408
			frame_aspect_ratio = enc->aspect_ratio;
/*#elif FFMPEG_VERSION_INT == 0x000409
            frame_aspect_ratio = av_q2d(enc->sample_aspect_ratio) * enc->width / enc->height;
*/
#else
    #error "there is a current unsupported ffmpeg version or FFMPEG_VERSION_INT isnt defined"
#endif
			frame_pix_fmt = enc->pix_fmt;
			rfps = ic->streams[i]->r_frame_rate;
			rfps_base = ic->streams[i]->r_frame_rate_base;
			enc->workaround_bugs = FF_BUG_AUTODETECT;
			enc->error_resilience = 2;
			enc->error_concealment = 3;
			enc->idct_algo = 0;
#ifdef DEBUG
			enc->debug = 1;
#else
			enc->debug = 0;
#endif

			//assert (enc->frame_rate_base == rfps_base);	// should be true for now
			if (enc->frame_rate != rfps)
			{
				fprintf (stderr,
					 "\nSeems that stream %d comes from film source: %2.2f->%2.2f\n",
					 i,
					 (float) enc->frame_rate /
					 enc->frame_rate_base,
					 (float) rfps / rfps_base);
			}
			// update the current frame rate to match the stream frame rate 
			frame_rate = rfps;
			frame_rate_base = rfps_base;

			enc->rate_emu = 0;
		}
		else
		{
			fprintf (stderr,
				 "Unsupported codec_type for input stream %d\n",
				 i);
			return false;
		}

	}

	// dump the file content 
#ifdef DEBUG
	dump_format (ic, 0, "", 0);
#endif
	return ic;
}

static AVFormatContext *
opt_output_file (const char *filename)
{
	AVOutputFormat *file_oformat;
	AVStream *st;
	AVFormatContext *oc;
	int use_video, nb_streams;
	AVFormatParameters params, *ap = &params;

	oc = (AVFormatContext *) av_mallocz (sizeof (AVFormatContext));

	file_oformat = guess_format (NULL, filename, NULL);
	if (!file_oformat)
	{
		fprintf (stderr,
			 "Unable for find a suitable output format for '%s'\n",
			 filename);
		return NULL;
	}

	oc->oformat = file_oformat;


	use_video = file_oformat->video_codec != CODEC_ID_NONE;

	nb_streams = 0;
	if (use_video)
	{
		AVCodecContext *video_enc;

		st = (AVStream *) av_mallocz (sizeof (AVStream));
		if (!st)
		{
			fprintf (stderr, "Could not alloc stream\n");
			return NULL;
		}
		avcodec_get_context_defaults (&st->codec);

		video_enc = &st->codec;


		video_enc->codec_id = file_oformat->video_codec;

		video_enc->bit_rate = 200 * 1000;
		video_enc->bit_rate_tolerance = 4000 * 1000;
		video_enc->frame_rate = 25;
		video_enc->frame_rate_base = 1;

        if((keep_aspect_ratio & 2) == keep_aspect_ratio) { // Nur Höhe wurde definiert
    		video_enc->width = ((int)((float)frame_height * frame_aspect_ratio)) & 0x7ffffff8;
	       	video_enc->height = frame_height;
        } else if((keep_aspect_ratio & 1) == keep_aspect_ratio) { // Nur Weite wurde definiert
	       	video_enc->width = frame_width;
    		video_enc->height = ((int)((float)frame_width / frame_aspect_ratio)) & 0x7ffffff8;
        } else {
    		video_enc->width = frame_width;
	       	video_enc->height = frame_height;
        }

#if FFMPEG_VERSION_INT == 0x000408
		video_enc->aspect_ratio = frame_aspect_ratio;
/*#elif FFMPEG_VERSION_INT == 0x000409
        video_enc->sample_aspect_ratio = av_d2q(frame_aspect_ratio*frame_height/frame_width, 255);
*/
#else
    #error "there is a current unsupported ffmpeg version or FFMPEG_VERSION_INT isnt defined"
#endif
		video_enc->pix_fmt = frame_pix_fmt;

		video_enc->gop_size = 0;

		video_enc->mb_decision = FF_MB_DECISION_SIMPLE;


		video_enc->qmin = 2;
		video_enc->qmax = 31;
		video_enc->mb_qmin = 2;
		video_enc->mb_qmax = 31;
		video_enc->max_qdiff = 3;
		video_enc->qblur = 0.5;
		video_enc->qcompress = 0.5;
		video_enc->rc_eq = "tex^qComp";
#ifdef DEBUG
		video_enc->debug = 1;
#else
		video_enc->debug = 0;
#endif
		video_enc->rc_override_count = 0;

		video_enc->rc_max_rate = 0;
		video_enc->rc_min_rate = 0;
		video_enc->rc_buffer_size = 0;
		video_enc->rc_buffer_aggressivity = 1.0;
		video_enc->rc_initial_cplx = 0;
		video_enc->i_quant_factor = -0.8;
		video_enc->b_quant_factor = 1.25;
		video_enc->i_quant_offset = 0;
		video_enc->b_quant_offset = 1.25;
		video_enc->dct_algo = 0;
		video_enc->idct_algo = 0;
		video_enc->strict_std_compliance = 0;

		video_enc->me_method = ME_EPZS;



		oc->streams[nb_streams] = st;
		nb_streams++;
	}

	oc->nb_streams = nb_streams;

	if (!nb_streams)
	{
		fprintf (stderr, "No video streams available\n");
		return NULL;
	}


	strcpy (oc->filename, filename);

	// check filename in case of an image number is expected 
	if (oc->oformat->flags & AVFMT_NEEDNUMBER)
	{
		if (filename_number_test (oc->filename) < 0)
		{
			print_error (oc->filename, AVERROR_NUMEXPECTED);
			return NULL;
		}
	}

	memset (ap, 0, sizeof (*ap));
	ap->image_format = NULL;
	if (av_set_parameters (oc, ap) < 0)
	{
		fprintf (stderr, "%s: Invalid encoding parameters\n",
			 oc->filename);
		return NULL;
	}

	return oc;
}


bool
decode (const std::string & szOutfile, const tPackedList & packed, int width,
	int height)
{
	AVFormatContext *ifc;
	AVFormatContext *ofc;

	av_register_all ();

	if (NULL == (ifc = opt_input_file (packed)))
		return false;
    
    if (width != -1 || height != -1) {
    	if (width != -1) {
    		frame_width = width;
            keep_aspect_ratio |= 1;
        } else {
            keep_aspect_ratio &= ~1;
        }

    	if (height != -1) {
    		frame_height = height;
            keep_aspect_ratio |= 2;
        } else {
            keep_aspect_ratio &= ~2;
        }
    }

	if (NULL == (ofc = opt_output_file (szOutfile.c_str ())))
	{
		fprintf (stderr, "Must supply at least one output file\n");
		return false;
	}


	bool bRet = (0 == av_encode (ofc, ifc, packed));

	/* close files */
	{
		/* maybe av_close_output_file ??? */
		AVFormatContext *s = ofc;
		int j;
		for (j = 0; j < s->nb_streams; ++j)
			av_free (s->streams[j]);
		av_free (s);
	}
	av_close_input_file (ifc);

	av_free_static ();

	return bRet;
}
