/*
 * Simple program to grab images from VDR Recording
 *
 * Copyright (c) 2005 Andreas Brachold
 * 
 * based on vdrsync.pl : Copyright (c) 2005 Peter Sebbel 
 *
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.  
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <iostream>
#include <fstream>
#include <sstream>
#include <vector>
#include <algorithm>
#include "ffm.h"



struct tGOP
{
	int n;
	union uIndex
	{
		struct tIndex
		{
			int offset;
			unsigned char type;
			unsigned char number;
			short reserved;
		} f;
		char rawdata[8];
	} u;
	int s;
	  tGOP ()
	{
	};
	tGOP (int nFrame)
	{
		n = nFrame;
	}
};

std::istream & operator >> (std::istream & i, tGOP & x)
{
	i.read ((x.u.rawdata), sizeof (x.u.rawdata));	// Read all data in one step
	return i;
}

#ifdef DEBUG
std::ostream & operator << (std::ostream & o, const tGOP & x)
{
	o << "Frame    : " << x.n;
	o << " Offset   : " << x.u.f.offset;
	o << " Type     : " << (int) x.u.f.type;
	o << " Number   : " << (int) x.u.f.number;
	o << " Reserved : " << x.u.f.reserved;
	return o;
}
#endif // DEBUG


bool
ReadIndexFile (const std::string & szFolder, std::vector < int >&nFrames,
	       std::vector < tGOP > &nGOP)
{
	if (szFolder.empty ())
	{
		std::cerr << "Missing VDR-recording folder" << std::endl;
		return false;
	}

	if (nFrames.empty ())
	{
		std::cerr << "Missing wanted frames" << std::endl;
		return false;
	}

	std::stringstream ss;
	ss << szFolder;
	if ('/' != *szFolder.rbegin ())
		ss << '/';
	ss << "index.vdr";
	ss << std::ends;
	std::string szFile (ss.str ());
	try
	{
		std::ifstream f (szFile.c_str (),
				 std::ifstream::in | std::ifstream::binary);
		
if (f.fail ())
		{
			std::cerr << "Can't open file : " << szFile << std::
				endl;
			return false;
		}
		std::vector < int >::iterator i = nFrames.begin ();
		std::vector < int >::iterator e = nFrames.end ();
		for (; i != e && f.good (); ++i)
		{
			tGOP x ((*i) - 1);
			f.seekg (x.n * 8, std::ifstream::beg);
			do
			{
				++x.n;
				if (!f.good ())
				{
					std::cerr <<
						"Seek behind end of file : ";
					std::cerr << szFile;
					std::cerr << " at Frame ";
					std::cerr << x.n << std::endl;
					return false;
				}
				f >> x;
			}
			while (x.u.f.type != 1);	// found I-Frame

			nGOP.push_back (x);	//Remember I-Frame

			do
			{
				++x.n;
				if (!f.good ())
				{
					std::cerr <<
						"Seek behind end of file : ";
					std::cerr << szFile;
					std::cerr << " at Frame ";
					std::cerr << x.n << std::endl;
					return false;
				}
				f >> x;

			}
			while (f.good () && x.u.f.type != 1);	// Build IBBP..I -> break on next I-Frame
			nGOP.push_back (x);	//Remember I-Frame
		}
		f.close ();
		return true;
	}
	catch (...)
	{
		std::cerr << "Something fail at read " << szFile << std::endl;
		return false;
	}
}

bool
ReadRecordings (const std::string & szFolder, const std::string & szOutFolder,
		std::vector < tGOP > &nGOP, int width, int height)
{
	if (szFolder.empty ())
	{
		std::cerr << "Missing VDR-recording folder" << std::endl;
		return false;
	}

	if (nGOP.empty ())
	{
		std::cerr << "Missing wanted GOP" << std::endl;
		return false;
	}

	std::vector < tGOP >::iterator i = nGOP.begin ();
	std::vector < tGOP >::iterator e = nGOP.end ();

	int nPicture = 1;
	int nOffset = 0;
	for (; i != e; ++i, ++nPicture)
	{
		int nFiles = (*i).u.f.number;
		std::stringstream ss;
		ss << szFolder;
		if ('/' != *szFolder.rbegin ())
			ss << '/';
		ss.fill ('0');
		ss.width (3);
		ss << nFiles;
		ss << ".vdr";
		ss << std::ends;
		std::string szFile (ss.str ());
		try
		{
			struct stat64 ds;
			if (stat64 (szFile.c_str (), &ds))
			{
				std::cerr << "Can't stat file : " << szFile <<
					std::endl;
				return false;
			}
			// Offset bigger then current 00x.vdr files, goto next
			if ((*i).u.f.offset >= (ds.st_size))
			{
				std::cerr <<
					"Offset bigger than current file : "
					<< szFile << std::endl;
				return false;
			}

			std::ifstream f (szFile.c_str (),
					 std::ifstream::in | std::ifstream::
					 binary);
			
            if (f.fail ())
			{
				std::cerr << "Can't open file : " << szFile <<
					std::endl;
				return false;
			}
			int nStart = (*i).u.f.offset;
			f.seekg (nStart - nOffset, std::ifstream::beg);	// Seek to I-Frame
			if (i == e)
			{
				std::cerr << "Missing second frame for : " <<
					szFile << std::endl;
				return false;
			}
			++i;
			// read GOP
			int nSize = (*i).u.f.offset - nStart;
			char *pMem = NULL;
			if (nSize <= 0
			    || nSize >= (100000 * 15)
			    || NULL == (pMem = new char[nSize]))
			{
				std::cerr << "Can't alloc memory for file : "
					<< szFile << std::endl;
				std::cerr << "want :" << nSize << " bytes" <<
					std::endl;
				return false;
			}
			f.read (pMem, nSize);
/* 
00 00 01 // Startcode
e4       // Streamtype
07 fa    // Length => 2042
84 c0    // notused
0b       // header_length
*/
			static const unsigned char packetMagic[3] =
				{ 0x00, 0x00, 0x01 };
			if (0 != memcmp (pMem, packetMagic, 3))
			{
				std::cerr <<
					"No valid magic found, at current packet, for file : "
					<< szFile << std::endl;
				std::cerr.fill ('0');
				std::cerr.width (2);
				std::cerr << "found : 0x" << std::
					hex << (unsigned int) (*(pMem + 0) &
							       0xFF) << std::
					hex << (unsigned int) (*(pMem + 1) &
							       0xFF) << std::
					hex << (unsigned int) (*(pMem + 2) &
							       0xFF) <<
					" streamcode : 0x";
				std::cerr.fill ('0');
				std::cerr.width (2);
				std::cerr << (unsigned int) (*(pMem + 3)) <<
					std::endl;
				delete[]pMem;
				return false;
			}
			int nOffset = 0;
			int nPackets = 0;

			std::stringstream so;
			so << szOutFolder;
			if ('/' != *szOutFolder.rbegin ())
				so << '/';
			so.fill ('0');
			so.width (8);
			so << (*i).n;
			so << "_%2d";
			so << ".jpg";
			so << std::ends;
			std::string szOut (so.str ());

#ifdef DEBUG
			std::stringstream sd;
			sd << szOutFolder;
			if ('/' != *szOutFolder.rbegin ())
				sd << '/';
			sd.fill ('0');
			sd.width (4);
			sd << (*i).n;
			sd << ".mpv";
			sd << std::ends;
			std::ofstream fo (sd.str ().c_str (),
					  std::ofstream::out | std::ofstream::
					  binary);
#endif //DEBUG
			tPackedList packed;

			while (nOffset < nSize)
			{
				int pLength =
					(((unsigned char)
					  *(pMem + nOffset + 4)) << 8) +
					(((unsigned char)
					  *(pMem + nOffset + 5)) << 0);
				int hLength =
					(((unsigned char)
					  *(pMem + nOffset + 8)) << 0);
				unsigned char cStream =
					(unsigned char) *(pMem + nOffset + 3);
				if ((cStream & 0xe0) == 0xe0)
				{	// Check for Stream 0xE0-0xEF
					// Store payload
					// |Header|......PaketLength| 
					unsigned char *pBuffer =
						(unsigned char *) pMem +
						nOffset + 9 + hLength;
					unsigned int nWrite =
						(pLength + 6) - (9 + hLength);

					packed.push_back (tPacked (pBuffer, nWrite));	//store startpoint and packed length
#ifdef DEBUG
					fo.write ((char *) pBuffer, nWrite);
#endif //DEBUG

					++nPackets;
				}
				nOffset += (pLength + 6);


				if (nOffset <= 0)	// avoid negativ seek
					break;
			}
			if (nPackets == 0 && nOffset >= nSize)
			{
				std::cerr <<
					"No valid packet found, for file : "
					<< szFile << std::endl;
				delete[]pMem;
				return false;
			}

#ifdef DEBUG
			{
				tPackedList::const_iterator i =
					packed.begin ();
				tPackedList::const_iterator e = packed.end ();
				for (; i != e; ++i)
				{
					std::cout << "Adress : 0x";
					std::cout.width (8);
					std::cout.fill ('0');
					std::cout << std::
						hex << (unsigned int) (void
								       *)
						(*i).
						pPacked << " Size : " << (*i).
						nPackedLength << std::endl;
				}
			}
#endif //DEBUG
			bool bRet = decode (szOut, packed, width, height);
			delete[]pMem;
			if (!bRet)
			{
				std::cerr << "decode failed, for file : " <<
					szFile << std::endl;
				return false;
			}
		}
		catch (...)
		{
			std::cerr << "Something fail at read " << szFile <<
				std::endl;
			return false;
		}
	}
	return true;
}

void
help (int argc, char *argv[])
{
	std::cerr << "Usage: " << argv[0] << std::endl;
	std::cerr << "            -r recordings    {VDR recording folder}" <<	std::endl;
	std::cerr << "            -f frame         {Wanted Frame 1/25s}" <<	std::endl;
	std::cerr << "            -o outdirectory  {output folder }" << std::endl;
	std::cerr << "            -x 160           {scaled width of output image}"	<< std::endl;
	std::cerr << "            -y 120           {scaled height of output image}"	<< std::endl;
	std::cerr << std::endl << std::endl;
}

int
option (int argc, char *argv[], const char opt, bool bParam,
	std::string & pararm, int n = 1)
{
	for (int i = n; i < argc; ++i)
	{
		if (argv[i] && strlen (argv[i]) > 1)
		{
			switch (*(argv[i] + 0))
			{
			case '-':
			{
				if (*(argv[i] + 1) == opt)
				{
					if (!bParam)
						return i + 1;
					if (i + 1 < argc)
					{
						pararm = argv[i + 1];
						return i + 2;
					}
				}
			}
			}
		}
	}
	return -1;
}

int
main (int argc, char *argv[])
{
	int width = -1;
	int height = -1;
	std::string szFolder, szOutFolder ("."), s;
	if (-1 == option (argc, argv, 'r', true, szFolder))
	{
		std::cerr << "missing vdr recording folder" << std::
			endl << std::endl;
		help (argc, argv);
		return 0;
	}

	if (-1 != option (argc, argv, 'x', true, s))
	{
		width = ((atoi (s.c_str ()) >> 1) << 1);	// Only /2
	}
	if (-1 != option (argc, argv, 'y', true, s))
	{
		height = ((atoi (s.c_str ()) >> 1) << 1);
	}
	option (argc, argv, 'o', true, szOutFolder);

	std::vector < int >nFrame;
	int n = 0;
	std::string f;
	
	while (-1 != (n = option (argc, argv, 'f', true, f, n)))
	{
		nFrame.push_back (atoi (f.c_str ()));
	}
	if (nFrame.empty ())
	{
		std::cerr << "missing wantet frame(s)" << std::endl << std::
			endl;
		help (argc, argv);
		return 0;
	}


	std::sort (nFrame.begin (), nFrame.end ());
#ifdef DEBUG
	std::vector < int >::const_iterator i = nFrame.begin ();
	std::vector < int >::const_iterator e = nFrame.end ();
	for (; i != e; ++i)
	{
		std::cout << "Wanted Frames : " << (*i) << std::endl;
	}
#endif //DEBUG

	std::vector < tGOP > nGOP;
	if (ReadIndexFile (szFolder, nFrame, nGOP))
	{
#ifdef DEBUG
		std::vector < tGOP >::iterator i = nGOP.begin ();
		std::vector < tGOP >::iterator e = nGOP.end ();
		for (; i != e; ++i)
		{
			std::cout << (*i) << std::endl;
		}
#endif //DEBUG
		if (ReadRecordings
		    (szFolder, szOutFolder, nGOP, width, height))
			return 1;	// Success
	}
	return 0;
}
